unit FrmMain;

//  ////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2006 Jan Punter
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  ////////////////////////////////////////////////////////////////////////////
//
//    lupd : 2006-03-25
//
//  ////////////////////////////////////////////////////////////////////////////


interface

uses

  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, IniFiles, ExtCtrls, Spin, HTTPApp, HTTPProd,
  StrUtils, BitHose, Data, Knowledge, Globals;

type

  TPatchData = array of byte;


  TCalculator = function( aPatchData: TPatchData; aStart, anEnd: Integer): Integer of object;

  TFormMain = class(TForm)
    OpenDialog: TOpenDialog;
    PanelTop: TPanel;
    BitBtnDoIt: TBitBtn;
    PanelInfo: TPanel;
    MemoDebug: TMemo;
    CheckBoxDumpBinary: TCheckBox;
    BitBtnMakeHtml: TBitBtn;
    PageProducer_modules: TPageProducer;
    PageProducer_page: TPageProducer;
    PageProducer_module: TPageProducer;
    PageProducer_input: TPageProducer;
    PageProducer_output: TPageProducer;
    PageProducer_param: TPageProducer;
    PageProducer_mode: TPageProducer;
    PageProducer_index: TPageProducer;
    PageProducer_index_item: TPageProducer;
    CheckBoxDumpKnowledge: TCheckBox;
    PanelAddress: TPanel;
    EditPatchName: TEdit;
    BitBtnDumpParameterKnowledge: TBitBtn;
    SpeedButton3: TSpeedButton;
    BitBtnBrowseInputFile: TBitBtn;
    Panel1: TPanel;
    SpeedButton1: TSpeedButton;
    SpeedButton2: TSpeedButton;
    BitBtnexit: TBitBtn;
    PageProducer_paramknowledge: TPageProducer;
    PageProducer_pknheader: TPageProducer;
    procedure BitBtnexitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure EditPatchNameChange(Sender: TObject);
    procedure BitBtnBrowseInputFileClick(Sender: TObject);
    procedure BitBtnDoItClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormDestroy(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure CheckBoxDumpBinaryClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure BitBtnMakeHtmlClick(Sender: TObject);
    procedure PageProducerHTMLTag(Sender: TObject; Tag: TTag;
      const TagString: String; TagParams: TStrings;
      var ReplaceText: String);
    procedure CheckBoxDumpKnowledgeClick(Sender: TObject);
    procedure BitBtnDumpParameterKnowledgeClick(Sender: TObject);
    procedure SpeedButton3Click(Sender: TObject);
    procedure FormResize(Sender: TObject);
  private
    FPatch                  : TPatch;
    FDumpData               : TStringList;
    FPatchName              : string;
    FFontSize               : Integer;
    FDumpBinary             : Boolean;
    FDumpKnowledge          : Boolean;
    FPosition               : Integer;
    FModuleKnowledgeTable   : TModuleKnowledgeTable;
    FParamKnowledgeTable    : TParamKnowledgeTable;
    FModuleKnowledgeName    : string;
    FParamKnowledgeName     : string;
    FAllowParamKnowledgDump : Boolean;
  private
    FCurrentPage            : TKnowledgePage;
    FCurrentKnowledge       : TModuleKnowledge;
    FCurrentInput           : string;
    FCurrentOutput          : string;
    FCurrentParam           : string;
    FCurrentMode            : string;
    FCurrentInputType       : string;
    FCurrentOutputType      : string;
    FCurrentParamRange      : string;
    FCurrentModeRange       : string;
    FCurrentParamType       : string;
    FCurrentModeType        : string;
    FCurrentInputId         : Integer;
    FCurrentOutputId        : Integer;
    FCurrentParamId         : Integer;
    FCurrentModeId          : Integer;
    FCurrentParamKnowledge  : TParamKnowledge;
  private
    procedure   SetModuleKnowledgeTable( aValue: TModuleKnowledgeTable);
    procedure   SetParamKnowledgeTable ( aValue: TParamKnowledgeTable );
    procedure   SetPatchName    ( const aValue: string);
    procedure   SetFontSize     ( aValue: Integer);
    procedure   SetDumpBinary   ( aValue: Boolean);
    procedure   SetDumpKnowledge( aValue: Boolean);
  private
    procedure   BrowseForPatch;
    procedure   SaveIni;
    procedure   LoadIni;
    function    InfoPath    ( const aFileName: string): string;
    function    TemplatePath( const aFileName: string): string;
    procedure   InitPageProducersForHtml;
    procedure   InitKnowledge;
    procedure   ExportHtml( const aFileName: string);
    function    EncodeSpecials( const aValue: string): string;
    function    Substitute( const aTag: string; aParams: TStrings): string;
  private
    procedure   ClearMsgs;
    procedure   ShowMsg( const aMsg: string);
    procedure   ShowMsgFmt( const aFormat: string; const anArgs: array of const);
  private
    procedure   ParsePatch;
    procedure   MakeParamKnowledgeDump( Compact: Boolean);
    procedure   ExportKnowledge;
  private
    property    ModuleKnowledgeTable : TModuleKnowledgeTable read FModuleKnowledgeTable write SetModuleKnowledgeTable;
    property    ParamKnowledgeTable  : TParamKnowledgeTable  read FParamKnowledgeTable  write SetParamKnowledgeTable;
  private
    property    PatchName     : string  read FPatchName     write SetPatchName;
    property    FontSize      : Integer read FFontSize      write SetFontSize;
    property    DumpBinary    : Boolean read FDumpBinary    write SetDumpBinary;
    property    DumpKnowledge : Boolean read FDumpKnowledge write SetDumpKnowledge;
  end;

var
  FormMain: TFormMain;

implementation

{$R *.dfm}

// User area


//  private

    procedure   TFormMain.SetModuleKnowledgeTable( aValue: TModuleKnowledgeTable);
    begin
      if Assigned( FModuleKnowledgeTable)
      then FreeAndNil( FModuleKnowledgeTable);
      FModuleKnowledgeTable := aValue;
    end;


    procedure   TFormMain.SetParamKnowledgeTable( aValue: TParamKnowledgeTable);
    begin
      if Assigned( FParamKnowledgeTable)
      then FreeAndNil( FParamKnowledgeTable);
      FParamKnowledgeTable := aValue;
    end;


    procedure   TFormMain.SetPatchName( const aValue: string);
    begin
      if aValue <> FPatchName
      then begin
        FPatchName := avalue;
        EditPatchName.Text := aValue;
      end;
    end;


    procedure   TFormMain.SetFontSize( aValue: Integer);
    begin
      if aValue <> FFontSize
      then begin
        FFontSize := aValue;
        MemoDebug.Font.Size := aValue;
      end;
    end;


    procedure   TFormMain.SetDumpBinary( aValue: Boolean);
    begin
      if aValue <> FDumpBinary
      then begin
        FDumpBinary := aValue;
        CheckBoxDumpBinary.Checked := aValue;
      end;
    end;


    procedure   TFormMain.SetDumpKnowledge( aValue: Boolean);
    begin
      if aValue <> FDumpKnowledge
      then begin
        FDumpKnowledge := aValue;
        CheckBoxDumpKnowledge.Checked := aValue;
      end;
    end;


//  private


    procedure   TFormMain.BrowseForPatch;
    begin
      with OpenDialog
      do begin
        if FileName = ''
        then FileName := PatchName;
        if InitialDir = ''
        then InitialDir := ApplicationPath;
        if Execute
        then PatchName := FileName;
      end;
    end;


  const sdfs = 'defaults';


    procedure   TFormMain.SaveIni;
    begin
      with TMemIniFile.Create( IniFileName)
      do
        try
          EraseSection( sdfs);
          WriteString ( sdfs, 'PatchName'    , PatchName    );
          WriteInteger( sdfs, 'FontSize'     , FontSize     );
          WriteInteger( sdfs, 'Left'         , Left         );
          WriteInteger( sdfs, 'Top'          , Top          );
          WriteInteger( sdfs, 'Width'        , Width        );
          WriteInteger( sdfs, 'Height'       , Height       );
          WriteBool   ( sdfs, 'DumpBinary'   , DumpBinary   );
          WriteBool   ( sdfs, 'DumpKnowledge', DumpKnowledge);
        finally
          UpdateFile;
          Free;
        end;
    end;


    procedure   TFormMain.LoadIni;
    var
      L, T, W, H : Integer;
    begin
      with TMemIniFile.Create( IniFileName)
      do
        try
          PatchName     := ReadString ( sdfs, 'PatchName'    , PatchName    );
          FontSize      := ReadInteger( sdfs, 'FontSize'     , FontSize     );
          L             := ReadInteger( sdfs, 'Left'         , Left         );
          T             := ReadInteger( sdfs, 'Top'          , Top          );
          W             := ReadInteger( sdfs, 'Width'        , Width        );
          H             := ReadInteger( sdfs, 'Height'       , Height       );
          DumpBinary    := ReadBool   ( sdfs, 'DumpBinary'   , DumpBinary   );
          DumpKnowledge := ReadBool   ( sdfs, 'DumpKnowledge', DumpKnowledge);
        finally
          free;
        end;
      if L < 0
      then L := 0;
      if T < 0
      then T := 0;
      if W > Screen.Width
      then W := Screen.Width;
      if H > Screen.Height
      then H := Screen.Height;
      SetBounds( L, T, W, H);
    end;


    function    TFormMain.InfoPath( const aFileName: string): string;
    begin
      Result :=
        IncludeTrailingPathDelimiter(
          IncludeTrailingPathDelimiter(
            ApplicationPath
          ) + 'info'
        ) + aFileName
    end;


    function    TFormMain.TemplatePath( const aFileName: string): string;
    begin
      Result :=
        IncludeTrailingPathDelimiter(
          IncludeTrailingPathDelimiter(
            ApplicationPath
          ) + 'templates'
        ) + aFileName
    end;


    procedure   TFormMain.InitPageProducersForHtml;
    begin
      PageProducer_modules       .HTMLFile := TemplatePath( 't_modules.html'       );
      PageProducer_index         .HTMLFile := TemplatePath( 't_index.html'         );
      PageProducer_index_item    .HTMLFile := TemplatePath( 't_index_item.html'    );
      PageProducer_page          .HTMLFile := TemplatePath( 't_page.html'          );
      PageProducer_module        .HTMLFile := TemplatePath( 't_module.html'        );
      PageProducer_input         .HTMLFile := TemplatePath( 't_input.html'         );
      PageProducer_output        .HTMLFile := TemplatePath( 't_output.html'        );
      PageProducer_param         .HTMLFile := TemplatePath( 't_param.html'         );
      PageProducer_mode          .HTMLFile := TemplatePath( 't_mode.html'          );
      PageProducer_paramknowledge.HTMLFile := TemplatePath( 't_paramknowledge.html');
      PageProducer_pknheader     .HTMLFile := TemplatePath( 't_pkn_header.html'    );
    end;


    procedure   TFormMain.InitKnowledge;
    begin
      if
        FileExists( FModuleKnowledgeName) and
        FileExists( FParamKnowledgeName)
      then begin
        ShowMsg( 'Reading module knowledge ... ');
        ModuleKnowledgeTable := TModuleKnowledgeTable.Create( FModuleKnowledgeName);
        ShowMsg( 'Reading param knowledge ... ');
        ParamKnowledgeTable := TParamKnowledgeTable.Create( FParamKnowledgeName);
        ShowMsg( 'Creating paged knowledge ... ');
        CreatePagedKnowledge( ModuleKnowledgeTable);
      end
      else begin
        ShowMsgFmt( 'Can''t find moduleKnowledge file : ''%s'', or', [ FModuleKnowledgeName]);
        ShowMsgFmt( 'can''t find paramKnowledge file : ''%s'', operation aborted', [ FParamKnowledgeName]);
        Abort;
      end;
    end;


    procedure   TFormMain.ExportHtml( const aFileName: string);
    begin
      ClearMsgs;
      try
        InitKnowledge;
        InitPageProducersForHtml;
        ShowMsg( 'Creating HTML output ... ');
        StringToFile( aFileName, PageProducer_modules.Content);
      except
        on E: exception
        do ShowMsgFmt( 'Exception : %s', [ E.Message]);
      end;
      ShowMsg( 'Done');
    end;


    function    TFormMain.EncodeSpecials( const aValue: string): string;
    // Replaces [text] by <A href="#text">text</a>
    // Replaces {+-}  by some proper symbol
    // Replaces {00} by some proper symbol (infinity)
    var
      i     : Integer;
      S     : string;
      State : ( stNormal, stInBrackets);
    begin
      // Bracket stuff
      Result := '';
      State  := stNormal;
      for i := 1 to Length( aValue)
      do begin
        case state of
          stNormal :
            case aValue[ i] of
              '[' : begin
                  State := stInBrackets;
                  S     := '';
                end;
            else Result := Result + aValue[ i];
            end;
          stInBrackets :
            case aValue[ i] of
              ']' : begin
                  Result := Result + Format( '<a href="#%s">%s</a>', [ S, S]);
                  State  := stNormal;
                end;
              else S := S + aValue[ i];
            end;
        end;
      end;
      // +- and 00 stuff
      Result := AnsiReplaceStr( Result, '{+-}', '&plusmn;');
      Result := AnsiReplaceStr( Result, '{00}', '&infin;');
    end;


    function    TFormMain.Substitute( const aTag: string; aParams: TStrings): string;
    //
      function TagIs( const aValue: string): Boolean;
      begin
        Result := SameText( aTag, aValue);
      end;
    //
    var
      i, j : Integer;
    begin
      Result := '';
      if TagIs( 'index')
      then begin
        with PagedKnowledge do
        begin
          for i := 0 to PageCount - 1
          do begin
            FCurrentPage := Page[ i];
            Result := Result + PageProducer_index.Content;
            FCurrentPage := nil;
          end;
        end;
      end
      else if TagIs( 'index_items')
      then begin
        if Assigned( FCurrentPage)
        then begin
          with FCurrentPage do
          begin
            for i := 0 to KnowledgeCount - 1
            do begin
              FCurrentKnowledge := Knowledge[ i];
              Result := Result + PageProducer_index_item.Content;
              FCurrentKnowledge := nil;
            end;
          end;
        end
        else Result := '[index_items : FCurrentPage not assigned]';
      end
      else if TagIs( 'pages')
      then begin
        with PagedKnowledge
        do begin
          for i := 0 to PageCount - 1 do
          begin
            FCurrentPage := Page[ i];
            Result := Result + PageProducer_page.Content;
            FCurrentPage := nil;
          end;
        end;
      end
      else if TagIs( 'page_name')
      then begin
        if Assigned( FCurrentPage)
        then Result := HTMLEncode( FCurrentPage.PageName)
        else Result := '[page_name : FCurrentPage not assigned]';
      end
      else if TagIs( 'module_name')
      then begin
        if Assigned( FCurrentKnowledge)
        then Result := HTMLEncode( FCurrentKnowledge.ShortName)
        else Result := '[module_name : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'modules')
      then begin
        if Assigned( FCurrentPage)
        then begin
          with FCurrentPage do
          begin
            for i := 0 to KnowledgeCount - 1
            do begin
              FCurrentKnowledge := Knowledge[ i];
              Result := Result + PageProducer_module.Content;
              FCurrentKnowledge := nil;
            end;
          end;
        end
        else Result := '[module : FCurrentPage not assigned]'
      end
      else if TagIs( 'module_short_name')
      then begin
        if Assigned( FCurrentKnowledge)
        then Result := HTMLEncode( FCurrentKnowledge.ShortName)
        else Result := '[module_short_name : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_long_name')
      then begin
        if Assigned( FCurrentKnowledge)
        then Result := HTMLEncode( FCurrentKnowledge.LongName)
        else Result := '[module_long_name : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_id')
      then begin
        if Assigned( FCurrentKnowledge)
        then Result := IntToStr( FCurrentKnowledge.TypeId)
        else Result := '[module_id : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_page')
      then begin
        if Assigned( FCurrentKnowledge)
        then Result := HTMLEncode( FCurrentKnowledge.Page)
        else Result := '[module_page : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'page_index')
      then begin
        if Assigned( FCurrentKnowledge)
        then Result := HTMLEncode( IntToStr( FCurrentKnowledge.PageIndex))
        else Result := '[page_index : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_class_name')
      then begin
        // maybe todo : maybe make FCurrentKnowledge.CleanName ?
        //   instead of using RawUrlEncode( ShortName)
        if Assigned( FCurrentKnowledge)
        then Result := HTTPEncode( FCurrentKnowledge.ShortName)
        else Result := '[module_class_name : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_inputs')
      then begin
        if Assigned( FCurrentKnowledge)
        then begin
          with FCurrentKnowledge
          do begin
            for i := 0 to InputCount - 1
            do begin
              FCurrentInputId   := i;
              FCurrentInput     := InputName[ i];
              FCurrentInputType := InputType[ i];
              Result            := Result + PageProducer_input.Content;
              FCurrentInputType := '';
              FCurrentInput     := '';
              FCurrentInputId   := -1;
            end;
          end;
        end
        else Result := '[module_inputs : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_outputs')
      then begin
        if Assigned( FCurrentKnowledge)
        then begin
          with FCurrentKnowledge
          do begin
            for i := 0 to OutputCount - 1
            do begin
              FCurrentOutputId   := i;
              FCurrentOutput     := OutputName[ i];
              FCurrentOutputType := OutputType[ i];
              Result             := Result + PageProducer_output.Content;
              FCurrentOutputType := '';
              FCurrentOutput     := '';
              FCurrentOutputId   := -1;
            end;
          end;
        end
        else Result := '[module_outputs : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_params')
      then begin
        if Assigned( FCurrentKnowledge)
        then begin
          with FCurrentKnowledge
          do begin
            for i := 0 to ParamCount - 1
            do begin
              FCurrentParamId    := i;
              FCurrentParam      := ParamName [ i];
              FCurrentParamRange := ParamRange[ i];
              FCurrentParamType  := ParamType [ i];
              Result             := Result + PageProducer_param.Content;
              FCurrentParamType  := '';
              FCurrentParamRange := '';
              FCurrentParam      := '';
              FCurrentParamId    := -1;
            end;
          end;
        end
        else Result := '[module_params : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'module_modes')
      then begin
        if Assigned( FCurrentKnowledge)
        then begin
          with FCurrentKnowledge
          do begin
            for i := 0 to ModeCount - 1
            do begin
              FCurrentModeId    := i;
              FCurrentMode      := ModeName [ i];
              FCurrentModeRange := ModeRange[ i];
              FCurrentModeType  := ModeType [ i];
              Result            := Result + PageProducer_mode.Content;
              FCurrentModeType  := '';
              FCurrentModeRange := '';
              FCurrentMode      := '';
              FCurrentModeId    := -1;
            end;
          end;
        end
        else Result := '[module_modes : FCurrentKnowledge not assigned]';
      end
      else if TagIs( 'input_id')
      then Result := IntToStr( FCurrentInputId)
      else if TagIs( 'input_name')
      then Result := HTMLEncode( FCurrentInput)
      else if TagIs( 'input_type')
      then Result := HTMLEncode( FCurrentInputType)
      else if TagIs( 'output_id')
      then Result := IntToStr( FCurrentOutputId)
      else if TagIs( 'output_name')
      then Result := HTMLEncode( FCurrentOutput)
      else if TagIs( 'output_type')
      then Result := HTMLEncode( FCurrentOutputType)
      else if TagIs( 'param_id')
      then Result := IntToStr( FCurrentParamId)
      else if TagIs( 'param_name')
      then Result := HTMLEncode( FCurrentParam)
      else if TagIs( 'param_range')
      then Result := HTMLEncode( FCurrentParamRange)
      else if TagIs( 'param_type')
      then Result := HTMLEncode( FCurrentParamType)
      else if TagIs( 'mode_id')
      then Result := IntToStr( FCurrentModeId)
      else if TagIs( 'mode_name')
      then Result := HTMLEncode( FCurrentMode)
      else if TagIs( 'mode_range')
      then Result := HTMLEncode( FCurrentModeRange)
      else if TagIs( 'mode_type')
      then Result := HTMLEncode( FCurrentModeType)
      else if TagIs( 'now')
      then Result := FormatDateTime( 'yyyy-mm-dd hh:nn', Now)
      else if TagIs( 'm_time_stamp_source')
      then begin
        if Assigned( FModuleKnowledgeTable)
        then Result := FormatDateTime( 'yyyy-mm-dd hh:nn', FModuleKnowledgeTable.TimeStamp)
        else Result := '[m_time_stamp_source : FModuleKnowledgeTable not assigned]';
      end
      else if TagIs( 'm_os_version_source')
      then begin
        if Assigned( FModuleKnowledgeTable)
        then Result := FModuleKnowledgeTable.OsVersion
        else Result := '[m_os_version_source : FModuleKnowledgeTable not assigned]';
      end
      else if TagIs( 'm_os_build_source')
      then begin
        if Assigned( FModuleKnowledgeTable)
        then Result := IntToStr( FModuleKnowledgeTable.OsBuild)
        else Result := '[m_os_build_source : FModuleKnowledgeTable not assigned]';
      end
      else if TagIs( 'p_time_stamp_source')
      then begin
        if Assigned( FParamKnowledgeTable)
        then Result := FormatDateTime( 'yyyy-mm-dd hh:nn', FParamKnowledgeTable.TimeStamp)
        else Result := '[p_time_stamp_source : FParamKnowledgeTable not assigned]';
      end
      else if TagIs( 'p_os_version_source')
      then begin
        if Assigned( FParamKnowledgeTable)
        then Result := FParamKnowledgeTable.OsVersion
        else Result := '[p_os_version_source : FParamKnowledgeTable not assigned]';
      end
      else if TagIs( 'p_os_build_source')
      then begin
        if Assigned( FParamKnowledgeTable)
        then Result := IntToStr( FParamKnowledgeTable.OsBuild)
        else Result := '[p_os_build_source : FParamKnowledgeTable not assigned]';
      end
      else if TagIs( 'param_knowledge')
      then begin
        with ParamKnowledgeTable
        do begin
          // Params
          Result := Result + PageProducer_pknheader.Content;
          for i := 0 to ParamKnowledgeCount - 1 do
          begin
            FCurrentParamKnowledge := ParamKnowledge[ i];
            if FCurrentParamKnowledge.ParamType = ptParam
            then Result := Result + PageProducer_paramknowledge.Content;
            FCurrentParamKnowledge := nil;
          end;
          // Modes
          Result := Result + PageProducer_pknheader.Content;
          for i := 0 to ParamKnowledgeCount - 1 do
          begin
            FCurrentParamKnowledge := ParamKnowledge[ i];
            if FCurrentParamKnowledge.ParamType = ptMode
            then Result := Result + PageProducer_paramknowledge.Content;
            FCurrentParamKnowledge := nil;
          end;
          // Undefined's
          Result := Result + PageProducer_pknheader.Content;
          for i := 0 to ParamKnowledgeCount - 1 do
          begin
            FCurrentParamKnowledge := ParamKnowledge[ i];
            if FCurrentParamKnowledge.ParamType = ptUndefined
            then Result := Result + PageProducer_paramknowledge.Content;
            FCurrentParamKnowledge := nil;
          end;
        end;
      end
      else if TagIs( 'pkn_param_type')
      then begin
        if Assigned( FCurrentParamKnowledge)
        then Result := HTMLEncode( ParamTypeName( FCurrentParamKnowledge.ParamType))
        else Result := '[pkn_param_type : FCurrentParamKnowledge not asssigned]'
      end
      else if TagIs( 'pkn_range_type')
      then begin
        if Assigned( FCurrentParamKnowledge)
        then Result := HTMLEncode( FCurrentParamKnowledge.RangeType)
        else Result := '[pkn_range_type : FCurrentParamKnowledge not asssigned]'
      end
      else if TagIs( 'pkn_range')
      then begin
        if Assigned( FCurrentParamKnowledge)
        then Result := FCurrentParamKnowledge.Range
        else Result := '[pkn_range : FCurrentParamKnowledge not asssigned]'
      end
      else if TagIs( 'pkn_default_value')
      then begin
        if Assigned( FCurrentParamKnowledge)
        then Result := Format( '%d', [ FCurrentParamKnowledge.DefaultValue])
        else Result := '[pkn_default_value : FCurrentParamKnowledge not asssigned]'
      end
      else if TagIs( 'pkn_definitions')
      then begin
        if Assigned( FCurrentParamKnowledge)
        then begin
          Result := '';
          with FCurrentParamKnowledge
          do begin
            for j := 0 to Length( Definitions) - 1
            do
              Result :=
                Result +
                  Format(
                    '%s%s',
                    [
                      EncodeSpecials(
                        HTMLEncode(
                          Definitions[ j]
                        )
                      ),
                      IfThen(
                        j <> Length( Definitions) - 1,
                        '<br>',
                        ''
                      )
                    ]
                  );
          end;
          if Result = ''
          then Result := '&nbsp;'
        end
        else Result := '[pkn_definitions : FCurrentParamKnowledge not asssigned]'
      end
      else if TagIs( 'pkn_comments')
      then begin
        if Assigned( FCurrentParamKnowledge)
        then begin
          Result := EncodeSpecials( HTMLEncode( FCurrentParamKnowledge.Comments));
          if Result = ''
          then Result := '&nbsp;'
        end
        else Result := '[pkn_comments : FCurrentParamKnowledge not asssigned]'
      end
      else Result := Format( '[unhandled_tag ''%s'']', [ aTag]);
    end;


//  private


    procedure   TFormMain.ClearMsgs;
    begin
      MemoDebug.Clear;
      Application.ProcessMessages;
    end;


    procedure   TFormMain.ShowMsg( const aMsg: string);
    begin
      MemoDebug.Lines.Add( aMsg);
    end;


    procedure   TFormMain.ShowMsgFmt( const aFormat: string; const anArgs: array of const);
    begin
      ShowMsg( Format( aFormat, anArgs))
    end;


//  private


    procedure   TFormMain.ParsePatch;
    var
      aRes      : Boolean;
    begin
      FPosition := MemoDebug.SelStart;
      ClearMsgs;
      try
        InitKnowledge;
        FPatch.ModuleKnowledgeTable := ModuleKnowledgeTable;
        aRes := FPatch.LoadFromFile( PatchName);
        if aRes
        then begin
          ShowMsg( 'Dumping patch ... ');
          FPatch.DumpToStrings( FDumpData, 0, DumpBinary, DumpKnowledge);
          MemoDebug.Lines.Assign( FDumpData);
          with MemoDebug do
          begin
            SelStart  := FPosition;
            SelLength := 0;
          end;
        end
        else ShowMsg( 'Could not parse patch');
      except
        on E: exception
        do ShowMsgFmt( 'Exception : %s', [ E.Message]);
      end;
    end;


    // /////////////////////////////////////////////////////////////////////////
    // /////////////////////////////////////////////////////////////////////////
    //
    // Some ugliness ...

    procedure   TFormMain.MakeParamKnowledgeDump( Compact: Boolean);

      procedure InitResult( aStringList: TStringList);
      begin
        with aStringList
        do begin
          Add( IndFmt( 0, 'ParamKnowledge('       , [                                         ] ));
          Add( IndFmt( 1,   'TimeStamp( %s)'      , [ FormatDateTime( 'yyyy-mm-dd hh:nn', Now)] ));
          Add( IndFmt( 1,   'OSVersion( ''1.40'')', [                                         ] ));
          Add( IndFmt( 1,   'OSBuild( 266)'       , [                                         ] ));
        end;
      end;

      procedure FinishResult( aStringList: TStringList);
      begin
        with aStringList
        do begin
          Add( IndFmt( 0, ')', [] ));
        end;
      end;

      procedure AddResult(
        aStringList: TStringList;
        aParamType    : string;
        aRangeType    : string;
        aLowValue     : Integer;
        aHighValue    : Integer
      );
      begin
        with aStringList
        do begin
          Add( IndFmt( 1, 'ParamDef('                 , [               ] ));
          Add( IndFmt( 2,   'ParamType( %s)'          , [ aParamType    ] ));
          Add( IndFmt( 2,   'RangeType( ''%s'')'      , [ aRangeType    ] ));
          Add( IndFmt( 2,   'LowValue( %d)'           , [ aLowValue     ] ));
          Add( IndFmt( 2,   'HighValue( %d)'          , [ aHighValue    ] ));
          Add( IndFmt( 2,   'DefaultValue( 0)'        , [               ] ));
          Add( IndFmt( 2,   'Definitions( [])'        , [               ] ));
          Add( IndFmt( 2,   'Comments( '''')'         , [               ] ));
          Add( IndFmt( 1, ')'                         , [               ] ));
        end;
      end;

    var
      i             : Integer;
      j             : Integer;
      S             : string;
      aKnowledge    : TModuleKnowledge;
      aStringList   : TStringList;
      aParamType    : string;
      aRangeType    : string;
      aLowValue     : Integer;
      aHighValue    : Integer;
      aResult       : TStringList;

    begin

      ClearMsgs;

      if not FAllowParamKnowledgDump
      then begin
        ShowMsg( 'Paramerer knowledge dumping not allowed, aborting');
        Exit;
      end;

      InitKnowledge;
      ShowMsg( 'Make parameter knowledge dump ...');

      try
        aStringList := TStringList.Create;
        aResult     := TStringList.Create;
        try
          aStringList.Sorted     := True;
          aStringList.Duplicates := dupIgnore;
          InitResult( aResult);
          with ModuleKnowledgeTable do
          begin
            for i := 0 to ModuleKnowledgeCount - 1
            do begin
              aKnowledge := Knowledge[ i];
              with aKnowledge
              do begin
                for j := 0 to ParamCount - 1
                do begin
                  if Compact
                  then
                    S :=
                      Format(
                        'param | %s | %s',
                        [
                          ExtendStr( ParamType [ j], 20),
                          ExtendStr( ParamRange[ j],  8)
                        ]
                      )
                  else
                    S :=
                      Format(
                        'param | %s | %s | %s | %s.%s',
                        [
                          ExtendStr( ParamType [ j], 20),
                          ExtendStr( ParamRange[ j],  8),
                          ExtendStr( ParamName [ j], 14),
                          Page,
                          ShortName
                        ]
                      );
                  if aStringList.IndexOf( S) < 0
                  then begin
                    aStringList.Add( S);
                    SplitParamRange( ParamRange[ j], aLowValue, aHighValue);
                    aParamType := 'Param';
                    aRangeType := ParamType[ j];
                    AddResult(
                      aResult,
                      aParamType,
                      aRangeType,
                      aLowValue,
                      aHighValue
                    );
                  end;
                end;
                for j := 0 to ModeCount - 1
                do begin
                  if Compact
                  then
                    S :=
                      Format(
                        'mode  | %s | %s',
                        [
                          ExtendStr( ModeType [ j], 20),
                          ExtendStr( ModeRange[ j],  8)
                        ]
                      )
                  else
                    S :=
                      Format(
                        'mode  | %s | %s | %s | %s.%s',
                        [
                          ExtendStr( ModeType [ j], 20),
                          ExtendStr( ModeRange[ j],  8),
                          ExtendStr( ModeName [ j], 14),
                          Page,
                          ShortName
                        ]
                      );
                  if aStringList.IndexOf( S) < 0
                  then begin
                    aStringList.Add( S);
                    SplitParamRange( ModeRange[ j], aLowValue, aHighValue);
                    aParamType := 'Mode';
                    aRangeType := ModeType[ j];
                    AddResult(
                      aResult,
                      aParamType,
                      aRangeType,
                      aLowValue,
                      aHighValue
                    );
                  end;
                end;
              end;
            end;
          end;
          FinishResult( aResult);
          aStringList.SaveToFile( InfoPath( 'params.txt'));
          aResult    .SaveToFile( InfoPath( 'param_table.txt'));
        finally
          aStringList.Free;
          aResult    .Free;
          ShowMsg( 'done');
        end;
      except
        on E: exception
        do ShowMsgFmt( 'Exception : %s', [ E.Message]);
      end

    end;

    // /////////////////////////////////////////////////////////////////////////
    // /////////////////////////////////////////////////////////////////////////

    procedure   TFormMain.ExportKnowledge;
    begin
      ClearMsgs;
      try
        InitKnowledge;
        ShowMsg( 'Export knowledge ...');
        ModuleKnowledgeTable.DumpKnowledge( InfoPath( 'module_table.txt'));
      except
        on E: exception
        do ShowMsgFmt( 'Exception : %s', [ E.Message]);
      end;
      ShowMsg( 'Done');
    end;


// Delphi area



procedure TFormMain.BitBtnexitClick(Sender: TObject);
begin
  Close;
end;

procedure TFormMain.FormCreate(Sender: TObject);
begin

  FAllowParamKnowledgDump := False;
  BitBtnDumpParameterKnowledge.Visible := FAllowParamKnowledgDump;

  FPatch    := TPatch.Create( nil);
  FDumpData := TStringList.Create;
  FModuleKnowledgeName :=
    IncludeTrailingPathDelimiter(
      IncludeTrailingPathDelimiter( ApplicationPath) + 'info'
    ) + 'module_table.txt';
  FParamKnowledgeName :=
    IncludeTrailingPathDelimiter(
      IncludeTrailingPathDelimiter( ApplicationPath) + 'info'
    ) + 'param_table.txt';
  FDumpBinary    := True; DumpBinary    := False;
  FDumpKnowledge := True; DumpKnowledge := False;
  ClearMsgs;
  LoadIni;
end;

procedure TFormMain.EditPatchNameChange(Sender: TObject);
begin
  PatchName := EditPatchName.Text;
end;

procedure TFormMain.BitBtnBrowseInputFileClick(Sender: TObject);
begin
  BrowseForPatch;
end;

procedure TFormMain.BitBtnDoItClick(Sender: TObject);
begin
  ParsePatch;
end;

procedure TFormMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  SaveIni;
end;

procedure TFormMain.FormDestroy(Sender: TObject);
begin
  FreeAndNil( FPatch);
  FreeAndNil( FDumpData);
  ModuleKnowledgeTable := nil;
  ParamKnowledgeTable  := nil;
end;

procedure TFormMain.SpeedButton1Click(Sender: TObject);
begin
  if FontSize < 20
  then FontSize := FontSize + 1;
end;

procedure TFormMain.SpeedButton2Click(Sender: TObject);
begin
  if FontSize > 4
  then FontSize := FontSize - 1;
end;

procedure TFormMain.CheckBoxDumpBinaryClick(Sender: TObject);
begin
  DumpBinary := CheckBoxDumpBinary.Checked;
end;

procedure TFormMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Shift = []
  then begin
    case Key of
      VK_F2 : BrowseForPatch;
      VK_F9 : ParsePatch;
    end;
  end;
end;

procedure TFormMain.FormShow(Sender: TObject);
begin
  //
end;

procedure TFormMain.BitBtnMakeHtmlClick(Sender: TObject);
begin
  ExportHtml( InfoPath( 'module_table.html'));
end;

procedure TFormMain.PageProducerHTMLTag(Sender: TObject; Tag: TTag;
  const TagString: String; TagParams: TStrings; var ReplaceText: String);
begin
  ReplaceText := Substitute( TagString, TagParams);
end;

procedure TFormMain.CheckBoxDumpKnowledgeClick(Sender: TObject);
begin
  DumpKnowledge := CheckBoxDumpKnowledge.Checked;
end;

procedure TFormMain.BitBtnDumpParameterKnowledgeClick(Sender: TObject);
begin
  // too dangerous now ... going to edit it as a source ...
  MakeParamKnowledgeDump( True);
end;

procedure TFormMain.SpeedButton3Click(Sender: TObject);
begin
  ExportKnowledge;
end;

procedure TFormMain.FormResize(Sender: TObject);
begin
  EditPatchName.Width := PanelAddress.Width - EditPatchName.Left - 2;
end;

end.
